/*
 * Linux Support Services, Inc.  Wildcard T100P T1/PRI card Driver
 *
 * Ported on FreeBSD by
 *      Alexander Timoshenko <gonzo@portaone.com>, PortaOne software
 *
 * Based on previous works, designs, and architectures conceived 
 * written by Mark Spencer <markster@linux-support.net>
 *            Matthew Fredrickson <creslin@linux-support.net>
 *            William Meadows <wmeadows@linux-support.net>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: wct1xxp.c 16 2004-10-26 04:42:50Z riz $
 */

#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>   /* cdevsw struct */
#include <sys/uio.h>	/* uio struct */
#include <sys/malloc.h>
#include <sys/lock.h>
#include <sys/endian.h>
#include <sys/bus.h>  /* structs, prototypes for pci bus stuff */

#include <machine/bus.h>
#include <sys/rman.h>
#include <machine/resource.h>
#include <machine/clock.h>
#include <sys/sysctl.h>

#include <vm/vm.h>		/* vtophys */
#include <vm/pmap.h>	/* vtophys */

#if __FreeBSD_version < 500000
#include <pci/pcireg.h>
#include <pci/pcivar.h>
#else
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#endif

#if __FreeBSD_version < 500000
#include <sys/kobj.h>
#endif



#include "os.h"
#include "zaptel.h"

#include "wct1xxp_var.h"
#include "wct1xxp_reg.h"

#define WCT1_LOCK(sc) {};
#define WCT1_UNLOCK(sc) {};

#define WC_MAX_CARDS	32

/*
#define TEST_REGS
*/

/* Define to get more attention-grabbing but slightly more I/O using
   alarm status */
#define FANCY_ALARM


#define DELAY	0x0	/* 30 = 15 cycles, 10 = 8 cycles, 0 = 3 cycles */

#define WC_CNTL    	0x00
#define WC_OPER		0x01
#define WC_AUXC    	0x02
#define WC_AUXD    	0x03
#define WC_MASK0   	0x04
#define WC_MASK1   	0x05
#define WC_INTSTAT 	0x06

#define WC_DMAWS	0x08
#define WC_DMAWI	0x0c
#define WC_DMAWE	0x10
#define WC_DMARS	0x18
#define WC_DMARI	0x1c
#define WC_DMARE	0x20
#define WC_CURPOS	0x24

#define WC_SERC		0x2d
#define WC_FSCDELAY	0x2f

#define WC_USERREG	0xc0

#define WC_CLOCK	0x0
#define WC_LEDTEST	0x1
#define WC_VERSION	0x2

/* Offset between transmit and receive */
#define WC_OFFSET	4

#define BIT_CS		(1 << 7)
#define BIT_ADDR	(0xf << 3)

#define BIT_LED0	(1 << 0)
#define BIT_LED1	(1 << 1)
#define BIT_TEST	(1 << 2)

static char *chips[] =
{
	"DS2152",
	"DS21352",
	"DS21552",
	"Unknown Chip (3)",
	"DS2154",
	"DS21354",
	"DS21554",
	"Unknown Chip (7)",
};

static int chanmap_t1[] = 
{ 2,1,0,
  6,5,4,
  10,9,8,
  14,13,12,
  18,17,16,
  22,21,20,
  26,25,24,
  30,29,28 };

static int chanmap_e1[] = 
{ 2,1,0,
  7,6,5,4,
  11,10,9,8,
  15,14,13,12,
  19,18,17,16,
  23,22,21,20,
  27,26,25,24,
  31,30,29,28 };

#ifdef FANCY_ALARM
static int altab[] = {
0, 0, 0, 1, 2, 3, 4, 6, 8, 9, 11, 13, 16, 18, 20, 22, 24, 25, 27, 28, 29, 30, 31, 31, 32, 31, 31, 30, 29, 28, 27, 25, 23, 22, 20, 18, 16, 13, 11, 9, 8, 6, 4, 3, 2, 1, 0, 0, 
};
#endif

/* sysctl stuff */
extern struct sysctl_oid * zt_root;
static struct sysctl_oid * wct1xxp_root;
static struct sysctl_ctx_list wct1xxp_clist;

/* prototypes */
void wct1xxp_sysctl_init __P((void));
void wct1xxp_sysctl_destroy __P((void));
void wct1xxp_dma_map_addr __P((void *, bus_dma_segment_t *, int, int));
static int wct1xxp_attach __P((device_t dev));
static int wct1xxp_detach __P((device_t dev));

/* syscall handlers */
d_open_t wct1xxp_sys_open;
d_close_t wct1xxp_sys_close;
d_read_t wct1xxp_sys_read;

static char * card_configuration = 0;

/* Character device entry points */
static struct cdevsw wct1xxp_cdevsw = {
#if __FreeBSD_version >= 502103
	.d_version = D_VERSION,
#endif
	.d_name  =   "wct1xxp",
	.d_open  =   wct1xxp_sys_open,
	.d_close =   wct1xxp_sys_close,
	.d_read  =   wct1xxp_sys_read
};




struct wct1xxp_softc {
	struct pci_dev *dev;
	lock_t lock;
	int ise1;
	int num;
	/* Our offset for finding channel 1 */
	int offset;
#if __FreeBSD_version < 500000
	dev_t d;
#else
	struct cdev * d; /* need to keep it for destroy_dev() */
#endif
	
	char *variety;
	int intcount;
	int usecount;
	int clocktimeout;
	int sync;
	int dead;
	int blinktimer;
	int alarmtimer;
	int loopupcnt;
	int loopdowncnt;
	int miss;
	int misslast;
	int *chanmap;
#ifdef FANCY_ALARM
	int alarmpos;
#endif
	unsigned char ledtestreg;
	unsigned char outbyte;
	unsigned long ioaddr;
	unsigned short canary;
	/* T1 signalling */
	unsigned char txsiga[3];
	unsigned char txsigb[3];
	struct resource *		wc_base_addr;
	bus_space_tag_t 		wc_base_addr_tag;
	bus_space_handle_t 		wc_base_addr_handle;
	int						wc_ioport_rid;

	struct wct1_data		wc_ldata;

	struct resource			* wc_irq;
	void *					irq_handler;
	bus_space_tag_t			wc_irq_bt;
	bus_space_handle_t		wc_irq_bh;


	unsigned char ec_chunk1[31][ZT_CHUNKSIZE];
	unsigned char ec_chunk2[31][ZT_CHUNKSIZE];
	unsigned char tempo[32];
	struct zt_span span;						/* Span */
	struct zt_chan chans[31];					/* Channels */
};


static struct wct1xxp_desc wct1xxp = 
	{ "Digium Wildcard T100P T1/PRI or E100P E1/PRA Board", 0 };

static struct wct1xxp_ident wct1xxp_ident_table[] = {
	/* {vendor, device, subverndor, card info }, */
	{ 0xe159, 0x0001, 0x6159,	&wct1xxp },
	{ 0, 0, 0, 0},
};

#define CANARY 0xca1e

int debug = 0;   /* doesnt do anything */

MALLOC_DECLARE(M_T1);
MALLOC_DEFINE(M_T1, "wct1xxp", "wct1xxp interface data structures");

static struct wct1xxp_softc *cards[WC_MAX_CARDS];


static inline void start_alarm(struct wct1xxp_softc *sc)
{
#ifdef FANCY_ALARM
	sc->alarmpos = 0;
#endif
	sc->blinktimer = 0;
}

static inline void stop_alarm(struct wct1xxp_softc *sc)
{
#ifdef FANCY_ALARM
	sc->alarmpos = 0;
#endif
	sc->blinktimer = 0;
}

static inline void __select_framer(struct wct1xxp_softc *sc, int reg)
{
	/* Top four bits of address from AUX 6-3 */
	sc->outbyte &= ~BIT_CS;
	sc->outbyte &= ~BIT_ADDR;
	sc->outbyte |= (reg & 0xf0) >> 1;
	WRITE1(WC_AUXD, sc->outbyte);
}

static inline void __select_control(struct wct1xxp_softc *sc)
{
	if (!(sc->outbyte & BIT_CS)) {
		sc->outbyte |= BIT_CS;
		WRITE1(WC_AUXD, sc->outbyte);
	}
}

static int t1xxp_open(struct zt_chan *chan)
{
	struct wct1xxp_softc *sc = chan->pvt;
	if (sc->dead)
		return -ENODEV;
	sc->usecount++;
	return 0;
}

static int __t1_get_reg(struct wct1xxp_softc *sc, int reg)
{
	unsigned char res;
	__select_framer(sc, reg);
	/* Get value */
	res = READ1(WC_USERREG + ((reg & 0xf) << 2));
	return res;
}

static int __t1_set_reg(struct wct1xxp_softc *sc, int reg, unsigned char val)
{
	__select_framer(sc, reg);
	/* Send address */
	WRITE1( WC_USERREG + ((reg & 0xf) << 2), val);
	return 0;
}

static int __control_set_reg(struct wct1xxp_softc *sc, int reg, 
	unsigned char val)
{
	__select_control(sc);
	WRITE1( WC_USERREG + ((reg & 0xf) << 2), val);
	return 0;
}

static int control_set_reg(struct wct1xxp_softc *sc, int reg, unsigned char val)
{
	int res;
	WCT1_LOCK(sc);
	res = __control_set_reg(sc, reg, val);
	WCT1_UNLOCK(sc);
	return res;
}

static int __control_get_reg(struct wct1xxp_softc *sc, int reg)
{
	unsigned char res;
	/* The following makes UTTERLY no sense, but what was happening
	   was that reads in some cases were not actually happening
	   on the physical bus. Why, we dunno. But in debugging, we found
	   that writing before reading (in this case to an unused position)
	   seems to get rid of the problem */
	__control_set_reg(sc, 3, 0x69); /* do magic here */
	/* now get the read byte from the Xilinx part */
	res = READ1(WC_USERREG + ((reg & 0xf) << 2));
	return res;
}

static int control_get_reg(struct wct1xxp_softc *sc, int reg)
{
	int res;
	WCT1_LOCK(sc);
	res = __control_get_reg(sc, reg);
	WCT1_UNLOCK(sc);
	return res;
}

static void t1xxp_release(struct wct1xxp_softc *sc)
{
	zt_unregister(&sc->span);
	printf("Freed a Wildcard\n");
}

static int t1xxp_close(struct zt_chan *chan)
{
	struct wct1xxp_softc *sc = chan->pvt;
	sc->usecount--;
	/* If we're dead, release us now */
	if (!sc->usecount && sc->dead) 
		t1xxp_release(sc);
	return 0;
}

static void t1xxp_enable_interrupts(struct wct1xxp_softc *sc)
{
	/* Clear interrupts */
	WRITE1(WC_INTSTAT, 0xff);
	/* Enable interrupts (we care about all of them) */
	WRITE1( WC_MASK0, 0x3c /* 0x3f */); 
	/* No external interrupts */
	WRITE1(WC_MASK1, 0x00);
}

static void t1xxp_start_dma(struct wct1xxp_softc *sc)
{
	/* Reset Master and TDM */
	WRITE1( WC_CNTL, DELAY | 0x0f);
	/*
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(1);
	*/
	tsleep(sc, PZERO | PCATCH, "waitintr", 2);
	WRITE1( WC_CNTL, DELAY | 0x01);
	WRITE1( WC_OPER, 0x01);
	if (debug) printf("Started DMA\n");
}

static void __t1xxp_stop_dma(struct wct1xxp_softc *sc)
{
	WRITE1( WC_OPER, 0x00);
}

static void __t1xxp_disable_interrupts(struct wct1xxp_softc *sc)	
{
	WRITE1( WC_MASK0, 0x00);
	WRITE1( WC_MASK1, 0x00);
}

static void __t1xxp_set_clear(struct wct1xxp_softc *sc)
{
	/* Setup registers */
	int x,y;
	unsigned char b;

	/* No such thing under E1 */
	if (sc->ise1) {
		printf("Can't set clear mode on an E1!\n");
		return;
	}

	for (x=0;x<3;x++) {
		b = 0;
		for (y=0;y<8;y++)
			if (sc->chans[x * 8 + y].sig & ZT_SIG_CLEAR)
				b |= (1 << y);
		__t1_set_reg(sc, 0x39 + x, b);
	}
}

static void t1xxp_t1_framer_start(struct wct1xxp_softc *sc)
{
	int i;
	char *coding, *framing;
	int alreadyrunning = sc->span.flags & ZT_FLAG_RUNNING;

	WCT1_LOCK(sc);

	/* Build up config */
	i = 0x20;
	if (sc->span.lineconfig & ZT_CONFIG_ESF) {
		coding = "ESF";
		i = 0x88;
	} else {
		coding = "SF";
	}
	if (sc->span.lineconfig & ZT_CONFIG_B8ZS) {
		framing = "B8ZS";
		i |= 0x44;
	} else {
		framing = "AMI";
	}
	__t1_set_reg(sc, 0x38, i);
	if (!(sc->span.lineconfig & ZT_CONFIG_ESF)) {
		/* 1c in FDL bit */
		__t1_set_reg(sc, 0x7e, 0x1c);
	} else {
		__t1_set_reg(sc, 0x7e, 0x00);
	}

	/* Set outgoing LBO */
	__t1_set_reg(sc, 0x7c, sc->span.txlevel << 5);

	printf("Using %s/%s coding/framing\n", coding, framing);
	if (!alreadyrunning) {
		/* Setup the clear channels */
		__t1xxp_set_clear(sc);

		/* Set LIRST bit to 1 */
		__t1_set_reg(sc, 0x0a, 0x80);
		WCT1_UNLOCK(sc);

		/* Wait 100ms to give plenty of time for reset */
		tsleep(sc, PZERO, "t1xxp_t1_framer_start", (hz / 10) + 1);

		WCT1_LOCK(sc);
		
		/* Reset LIRST bit and reset elastic stores */
		__t1_set_reg(sc, 0xa, 0x30);

		sc->span.flags |= ZT_FLAG_RUNNING;
	}
	WCT1_UNLOCK(sc);
}

static void t1xxp_e1_framer_start(struct wct1xxp_softc *sc)
{
	int i;
	char *coding, *framing;
	int alreadyrunning = sc->span.flags & ZT_FLAG_RUNNING;
	char *crcing = "";
	unsigned char ccr1, tcr1, tcr2;

	WCT1_LOCK(sc);

	/* Build up config */
	ccr1 = 0;
	tcr1 = 8;
	tcr2 = 0;
	if (sc->span.lineconfig & ZT_CONFIG_CCS) {
		coding = "CCS"; /* Receive CCS */
		ccr1 |= 8;
	} else {
		tcr1 |= 0x20;
		coding = "CAS";
	}
	if (sc->span.lineconfig & ZT_CONFIG_HDB3) {
		ccr1 |= 0x44;		/* TX/RX HDB3 */
		framing = "HDB3";
	} else {
		framing = "AMI";
	}
	if (sc->span.lineconfig & ZT_CONFIG_CRC4) {
		ccr1 |= 0x11;
		tcr2 |= 0x02;
		crcing = " with CRC4";
	}
	__t1_set_reg(sc, 0x12, tcr1);
	__t1_set_reg(sc, 0x13, tcr2);
	__t1_set_reg(sc, 0x14, ccr1);
	__t1_set_reg(sc, 0x18, 0x20);	/* 120 Ohm */
	
	
#if 0	/* XXX Does LBO Matter? XXX */
	/* Set outgoing LBO */
	__t1_set_reg(sc, 0x7c, sc->span.txlevel << 5);
#endif

	printf("Using %s/%s coding/framing%s 120 Ohms\n", coding, framing,crcing);
	if (!alreadyrunning) {
	
		__t1_set_reg(sc,0x1b,0x8a); /* CCR3: LIRST & TSCLKM */
		__t1_set_reg(sc,0x20,0x1b); /* TAFR */
		__t1_set_reg(sc,0x21,0x5f); /* TNAFR */
		__t1_set_reg(sc,0x40,0xb); /* TSR1 */
		for(i = 0x41; i <= 0x4f; i++) __t1_set_reg(sc,i,0x55);
		for(i = 0x22; i <= 0x25; i++) __t1_set_reg(sc,i,0xff);
		WCT1_UNLOCK(sc);

		/* Wait 100ms to give plenty of time for reset */
		tsleep(sc, PZERO, "t1xxp_e1_framer_start", (hz / 10) + 1);

		WCT1_LOCK(sc);
		
		__t1_set_reg(sc, 0x1b, 0x9a);	/* Set ESR */
		__t1_set_reg(sc, 0x1b, 0x82);	/* TSCLKM only now */

		/* Reset LIRST bit and reset elastic stores */

		sc->span.flags |= ZT_FLAG_RUNNING;
	}
	WCT1_UNLOCK(sc);
}

static int t1xxp_framer_sanity_check(struct wct1xxp_softc *sc)
{
	int res;
	int chipid;
	int x;

	/* Sanity check */
	WCT1_LOCK(sc);
	for (x=0x0;x<192;x++)
		__t1_set_reg(sc, x, 0);
	res = __t1_get_reg(sc, 0x0f);
	res = __t1_get_reg(sc, 0x0f);
	chipid = ((res & 0x80) >> 5) | ((res & 0x30) >> 4);
	sc->ise1 = (res & 0x80) ? (1 << 4) : 0;
	WCT1_UNLOCK(sc);

	printf("Framer: %s, Revision: %d (%s)\n", chips[chipid], res & 0xf, 
		sc->ise1 ? "E1" : "T1");

	return 0;
}

static int t1xxp_framer_hard_reset(struct wct1xxp_softc *sc)
{
	int x;

	WCT1_LOCK(sc);
	/* Initialize all registers to 0 */
	for (x=0x0;x<192;x++)
		__t1_set_reg(sc, x, 0);

	if (sc->ise1) {
		/* Set LOTCMC (switch to RCLCK if TCLK fails) */
		__t1_set_reg(sc, 0x1a, 0x04);
		
	 	/* RSYNC is an input */
		__t1_set_reg(sc, 0x10, 0x20);

		/* Rx elastic store enabled, 2.048 Mhz (in theory) */
		__t1_set_reg(sc, 0x11, 0x06);

		/* TSYNC is an input, Tsis mode */
		__t1_set_reg(sc, 0x12, 0x08);

		/* Tx elastic store enabled, 2.048 Mhz (in theory) */
		__t1_set_reg(sc, 0x1b, 0x82);


		
	} else {
		/* Full-on sync required for T1 */
		__t1_set_reg(sc, 0x2b, 0x08);
	 	/* RSYNC is an input */
		__t1_set_reg(sc, 0x2c, 0x08);

		/* Enable tx RBS bits */
		__t1_set_reg(sc, 0x35, 0x10);

		/* TSYNC is output */
		__t1_set_reg(sc, 0x36, 0x04);

		/* Tx and Rx elastic store enabled, 2.048 Mhz (in theory) */
		__t1_set_reg(sc, 0x37, 0x9c);

		/* Setup Loopup / Loopdown codes */
		__t1_set_reg(sc, 0x12, 0x22);
		__t1_set_reg(sc, 0x14, 0x80);
		__t1_set_reg(sc, 0x15, 0x80);
	}

	WCT1_UNLOCK(sc);
	return 0;
}

static int t1xxp_rbsbits(struct zt_chan *chan, int bits)
{
	struct wct1xxp_softc *sc = chan->pvt;
	int b,o;
	unsigned char mask;
	
	/* Byte offset */
	WCT1_LOCK(sc);
	if (sc->ise1) {
                if (chan->chanpos < 16) {
                       mask = ((bits << 4) | sc->chans[chan->chanpos - 1 + 16].txsig);
                        __t1_set_reg(sc, 0x40 + chan->chanpos, mask);
                }
		else if (chan->chanpos > 16) {
			mask = (bits | (sc->chans[chan->chanpos - 1 - 16].txsig << 4));
			__t1_set_reg(sc, 0x40 + chan->chanpos - 16, mask);
		}
		sc->chans[chan->chanpos - 1].txsig = bits;
	} else {
		b = (chan->chanpos - 1) / 8;
		o = (chan->chanpos - 1) % 8;

		mask = (1 << o);

		if (bits & ZT_ABIT) {
			/* Set A-bit */
			sc->txsiga[b] |= mask;
		} else {
			/* Clear A-bit */
			sc->txsiga[b] &= ~mask;
		}
		if (bits & ZT_BBIT) {
			/* Set B-bit */
			sc->txsigb[b] |= mask;
		} else {
			sc->txsigb[b] &= ~mask;
		}
		/* Output new values */
		__t1_set_reg(sc, 0x70 + b, sc->txsiga[b]);
		__t1_set_reg(sc, 0x73 + b, sc->txsigb[b]);
		__t1_set_reg(sc, 0x76 + b, sc->txsiga[b]);
		__t1_set_reg(sc, 0x79 + b, sc->txsigb[b]);
	}
	WCT1_UNLOCK(sc);
	return 0;
}

static int t1xxp_ioctl(struct zt_chan *chan, unsigned int cmd, caddr_t data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
}

static int t1xxp_startup(struct zt_span *span)
{
	struct wct1xxp_softc *sc = span->pvt;

	int i,alreadyrunning = span->flags & ZT_FLAG_RUNNING;

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++)
	{
		memset(sc->ec_chunk1[i],
			ZT_LIN2X(0,&span->chans[i]),ZT_CHUNKSIZE);
		memset(sc->ec_chunk2[i],
			ZT_LIN2X(0,&span->chans[i]),ZT_CHUNKSIZE);
	}

	/* Reset framer with proper parameters and start */
	if (sc->ise1)
		t1xxp_e1_framer_start(sc);
	else
		t1xxp_t1_framer_start(sc);

	if(debug)
		printf("Calling startup (flags is %d)\n", span->flags);

	if (!alreadyrunning) {
		/* Only if we're not already going */
		t1xxp_enable_interrupts(sc);
		t1xxp_start_dma(sc);
		span->flags |= ZT_FLAG_RUNNING;
	}
	return 0;
}

static int t1xxp_shutdown(struct zt_span *span)
{
	struct wct1xxp_softc *sc = span->pvt;

	WCT1_LOCK(sc);
	__t1xxp_stop_dma(sc);
	__t1xxp_disable_interrupts(sc);
	span->flags &= ~ZT_FLAG_RUNNING;
	WCT1_UNLOCK(sc);

	t1xxp_framer_hard_reset(sc);
	return 0;
}

static int t1xxp_maint(struct zt_span *span, int cmd)
{
	struct wct1xxp_softc *sc = span->pvt;
	int res = 0;
	WCT1_LOCK(sc);
	if (sc->ise1) {
		switch(cmd) {
		case ZT_MAINT_NONE:
			__t1_set_reg(sc,0xa8,0); /* no loops */
			break;
		case ZT_MAINT_LOCALLOOP:
			__t1_set_reg(sc,0xa8,0x40); /* local loop */
			break;
		case ZT_MAINT_REMOTELOOP:
			__t1_set_reg(sc,0xa8,0x80); /* remote loop */
			break;
		case ZT_MAINT_LOOPUP:
		case ZT_MAINT_LOOPDOWN:
		case ZT_MAINT_LOOPSTOP:
			res = -ENOSYS;
			break;
		default:
			printf("wct1xxp/E1: Unknown maint command: %d\n", cmd);
			res = -EINVAL;
			break;
		}
	} else {
		switch(cmd) {
	    case ZT_MAINT_NONE:
			__t1_set_reg(sc,0x19,0); /* no local loop */
			__t1_set_reg(sc,0x0a,0); /* no remote loop */
			break;
	    case ZT_MAINT_LOCALLOOP:
			__t1_set_reg(sc,0x19,0x40); /* local loop */
			__t1_set_reg(sc,0x0a,0); /* no remote loop */
			break;
	    case ZT_MAINT_REMOTELOOP:
			__t1_set_reg(sc,0x1e,0); /* no local loop */
			__t1_set_reg(sc,0x0a,0x40); /* remote loop */
			break;
	    case ZT_MAINT_LOOPUP:
			__t1_set_reg(sc,0x30,2); /* send loopup code */
			__t1_set_reg(sc,0x12,0x22); /* send loopup code */
			__t1_set_reg(sc,0x13,0x80); /* send loopup code */
			break;
	    case ZT_MAINT_LOOPDOWN:
			__t1_set_reg(sc,0x30,2); /* send loopdown code */
			__t1_set_reg(sc,0x12,0x62); /* send loopdown code */
			__t1_set_reg(sc,0x13,0x90); /* send loopdown code */
			break;
	    case ZT_MAINT_LOOPSTOP:
			__t1_set_reg(sc,0x30,0);	/* stop sending loopup code */
			break;
	    default:
			printf("wct1xxp/T1: Unknown maint command: %d\n", cmd);
			res = -EINVAL;
	   }
	}
	WCT1_UNLOCK(sc);
	return res;
}

static int t1xxp_chanconfig(struct zt_chan *chan, int sigtype)
{
	struct wct1xxp_softc *sc = chan->pvt;
	int alreadyrunning = chan->span->flags & ZT_FLAG_RUNNING;

	WCT1_LOCK(sc);

	if (alreadyrunning && !sc->ise1)
		__t1xxp_set_clear(sc);

	WCT1_UNLOCK(sc);
	return 0;
}

static int t1xxp_spanconfig(struct zt_span *span, struct zt_lineconfig *lc)
{
	struct wct1xxp_softc *sc = span->pvt;
	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	/* Do we want to SYNC on receive or not */
	sc->sync = lc->sync;
	/* If already running, apply changes immediately */
	if (span->flags & ZT_FLAG_RUNNING)
		return t1xxp_startup(span);
	return 0;
}

static int t1xxp_software_init(struct wct1xxp_softc *sc)
{
	int x;
	/* Find position */
	for (x=0;x<WC_MAX_CARDS;x++) {
		if (!cards[x]) {
			cards[x] = sc;
			break;
		}
	}
	if (x >= WC_MAX_CARDS)
		return -1;
	sc->num = x;
	sprintf(sc->span.name, "WCT1/%d", sc->num);
	sprintf(sc->span.desc, "%s Card %d", sc->variety, sc->num);
	sc->span.spanconfig = t1xxp_spanconfig;
	sc->span.chanconfig = t1xxp_chanconfig;
	sc->span.startup = t1xxp_startup;
	sc->span.shutdown = t1xxp_shutdown;
	sc->span.rbsbits = t1xxp_rbsbits;
	sc->span.maint = t1xxp_maint;
	sc->span.open = t1xxp_open;
	sc->span.close = t1xxp_close;
	if (sc->ise1)
		sc->span.channels = 31;
	else
		sc->span.channels = 24;
	sc->span.chans = sc->chans;
	sc->span.flags = ZT_FLAG_RBS;
	sc->span.linecompat = ZT_CONFIG_AMI | ZT_CONFIG_B8ZS | 
		ZT_CONFIG_D4 | ZT_CONFIG_ESF;
	sc->span.ioctl = t1xxp_ioctl;
	sc->span.pvt = sc;
	if (sc->ise1)
		sc->span.deflaw = ZT_LAW_ALAW;
	else
		sc->span.deflaw = ZT_LAW_MULAW;

	for (x=0;x<sc->span.channels;x++) {
		sprintf(sc->chans[x].name, "WCT1/%d/%d", sc->num, x + 1);
		sc->chans[x].sigcap = ZT_SIG_EM | ZT_SIG_CLEAR | ZT_SIG_EM_E1 | 
				      ZT_SIG_FXSLS | ZT_SIG_FXSGS | 
				      ZT_SIG_FXSKS | ZT_SIG_FXOLS | 
				      ZT_SIG_FXOGS | ZT_SIG_FXOKS | ZT_SIG_CAS | ZT_SIG_SF;
		sc->chans[x].pvt = sc;
		sc->chans[x].chanpos = x + 1;
	}

	if (zt_register(&sc->span, 0)) {
		printf("Unable to register span with zaptel\n");
		return -1;
	}

	return 0;
}

static inline void __handle_leds(struct wct1xxp_softc *sc)
{
	int oldreg;
	sc->blinktimer++;

	if (sc->span.alarms & (ZT_ALARM_RED | ZT_ALARM_BLUE)) {
		/* Red/Blue alarm */
#ifdef FANCY_ALARM
		if (sc->blinktimer == (altab[sc->alarmpos] >> 1)) {
			sc->ledtestreg = (sc->ledtestreg | BIT_LED1) & ~BIT_LED0;
			__control_set_reg(sc, WC_LEDTEST, sc->ledtestreg);
		}
		if (sc->blinktimer == 0xf) {
			sc->ledtestreg = sc->ledtestreg & ~(BIT_LED0 | BIT_LED1);
			__control_set_reg(sc, WC_LEDTEST, sc->ledtestreg);
			sc->blinktimer = -1;
			sc->alarmpos++;
			if (sc->alarmpos >= (sizeof(altab) / sizeof(altab[0])))
				sc->alarmpos = 0;
		}
#else
		if (sc->blinktimer == 160) {
			sc->ledtestreg = (sc->ledtestreg | BIT_LED1) & ~BIT_LED0;
			__control_set_reg(sc, WC_LEDTEST, sc->ledtestreg);
		} else if (sc->blinktimer == 480) {
			sc->ledtestreg = sc->ledtestreg & ~(BIT_LED0 | BIT_LED1);
			__control_set_reg(sc, WC_LEDTEST, sc->ledtestreg);
			sc->blinktimer = 0;
		}
#endif
	} else if (sc->span.alarms & ZT_ALARM_YELLOW) {
		/* Yellow Alarm */
		if (!(sc->blinktimer % 2)) 
			sc->ledtestreg = (sc->ledtestreg | BIT_LED1) & ~BIT_LED0;
		else
			sc->ledtestreg = (sc->ledtestreg | BIT_LED0) & ~BIT_LED1;
		__control_set_reg(sc, WC_LEDTEST, sc->ledtestreg);
	} else {
		/* No Alarm */
		oldreg = sc->ledtestreg;
		if (sc->span.maintstat != ZT_MAINT_NONE)
			sc->ledtestreg |= BIT_TEST;
		else
			sc->ledtestreg &= ~BIT_TEST;
		if (sc->span.flags & ZT_FLAG_RUNNING)
			sc->ledtestreg = (sc->ledtestreg | BIT_LED0) & ~BIT_LED1;
		else
			sc->ledtestreg = sc->ledtestreg & ~(BIT_LED0 | BIT_LED1);
		if (oldreg != sc->ledtestreg)
			__control_set_reg(sc, WC_LEDTEST, sc->ledtestreg);
	}
}

static void t1xxp_transmitprep(struct wct1xxp_softc *sc, int ints)
{
	volatile unsigned char *txbuf;
	int x,y;
	int pos;
	if (ints & 0x04 /* 0x01 */) {
		/* We just finished sending the first buffer, start filling it
		   now */
		txbuf = sc->wc_ldata.wc_writechunk;
	} else {
		/* Just finished sending second buffer, fill it now */
		txbuf = sc->wc_ldata.wc_writechunk + 32 * ZT_CHUNKSIZE;
	}

	zt_transmit(&sc->span);
	for (x=0;x<sc->offset;x++)
		txbuf[x] = sc->tempo[x];
	for (y=0;y<ZT_CHUNKSIZE;y++) {
		for (x=0;x<sc->span.channels;x++) {
			pos = y * 32 + sc->chanmap[x] + sc->offset;
			/* Put channel number as outgoing data */
			if (pos < 32 * ZT_CHUNKSIZE)
				txbuf[pos] = sc->chans[x].writechunk[y];
			else
				sc->tempo[pos - 32 * ZT_CHUNKSIZE] = sc->chans[x].writechunk[y];
		}
	}
}

static void t1xxp_receiveprep(struct wct1xxp_softc *sc, int ints)
{
	volatile unsigned char * rxbuf;
	volatile unsigned int *canary;
	int x;
	int y;
	unsigned int oldcan;
	if (ints & 0x04) {
		/* Just received first buffer */
		rxbuf = sc->wc_ldata.wc_readchunk;
		canary = (volatile unsigned int *)
			(sc->wc_ldata.wc_readchunk + ZT_CHUNKSIZE * 64 - 4);
	} else {
		rxbuf = sc->wc_ldata.wc_readchunk + ZT_CHUNKSIZE * 32;
		canary = (volatile unsigned int *)
			(sc->wc_ldata.wc_readchunk + ZT_CHUNKSIZE * 32 - 4);
	}

	oldcan = *canary;
	if (((oldcan & 0xffff0000) >> 16) != CANARY) {
		/* Check top part */
		if (debug) 
			printf("Expecting top %04x, got %04x\n", 
				CANARY, (oldcan & 0xffff0000) >> 16);
		sc->span.irqmisses++;
	} else if ((oldcan & 0xffff) != ((sc->canary - 1) & 0xffff)) {
		if (debug) 
			printf("Expecting bottom %d, got %d\n", 
				sc->canary - 1, oldcan & 0xffff);
		sc->span.irqmisses++;
	}
	for (y=0;y<ZT_CHUNKSIZE;y++) {
		for (x=0;x<sc->span.channels;x++) {
			/* XXX Optimize, remove * and + XXX */
			/* Must map received channels into appropriate data */
			sc->chans[x].readchunk[y] = 
				rxbuf[32 * y + ((sc->chanmap[x] + WC_OFFSET + sc->offset) & 0x1f)];
		}
		if (!sc->ise1) {
			for (x=3;x<32;x+=4) {
				if (rxbuf[32 * y + ((x + WC_OFFSET) & 0x1f)] == 0x7f) {
					if (sc->offset != (x-3)) {
						/* Resync */
						control_set_reg(sc, WC_CLOCK, 0x02 | sc->sync | sc->ise1);
						sc->clocktimeout = 100;
#if 1
						if (debug) 
							printf("T1: Lost our place, resyncing (offset==%d, %d)\n", sc->offset, x);
#endif
					}
				}
			}
		} else {
			if (!sc->clocktimeout && !sc->span.alarms) {
				if ((rxbuf[32 * y + ((3 + WC_OFFSET + sc->offset) & 0x1f)] & 0x7f) != 0x1b) {
					if (sc->miss) {
						if (debug) printf("Double miss (%d, %d)...\n", sc->misslast, rxbuf[32 * y + ((3 + WC_OFFSET + sc->offset) & 0x1f)]);
						control_set_reg(sc, WC_CLOCK, 0x02 | sc->sync | sc->ise1);
						sc->clocktimeout = 100;
					} else {
						sc->miss = 1;
						sc->misslast = rxbuf[32 * y + ((3 + WC_OFFSET + sc->offset) & 0x1f)];
					}
				} else {
					sc->miss = 0;
				}
			} else {
				sc->miss = 0;
			}
		} 
	}
	/* Store the next canary */
	canary = (volatile unsigned int *)(rxbuf + ZT_CHUNKSIZE * 32 - 4);
	*canary = (sc->canary++) | (CANARY << 16);
	for (x=0;x<sc->span.channels;x++) {
		zt_ec_chunk(&sc->chans[x], sc->chans[x].readchunk, 
			sc->ec_chunk2[x]);
		memcpy(sc->ec_chunk2[x],sc->ec_chunk1[x],ZT_CHUNKSIZE);
		memcpy(sc->ec_chunk1[x],sc->chans[x].writechunk,ZT_CHUNKSIZE);
	}
	zt_receive(&sc->span);

	bus_dmamap_sync(sc->wc_ldata.wc_read_tag,
		sc->wc_ldata.wc_read_dmamap, BUS_DMASYNC_POSTWRITE | BUS_DMASYNC_POSTREAD);
}

static void __t1xxp_check_sigbits(struct wct1xxp_softc *sc, int x)
{
	int a,b,i,y,rxs;

	if (sc->ise1) {
		/* Read 5 registers at a time, loading 10 channels at a time */
		for (i = (x * 5); i < (x * 5) + 5; i++) {
			a = __t1_get_reg(sc, 0x31 + i);
			/* Get high channel in low bits */
			rxs = (a & 0xf);
			if (!(sc->chans[i+16].sig & ZT_SIG_CLEAR)) {
				if (sc->chans[i+16].rxsig != rxs)
					zt_rbsbits(&sc->chans[i+16], rxs);
			}
			rxs = (a >> 4) & 0xf;
			if (!(sc->chans[i].sig & ZT_SIG_CLEAR)) {
				if (sc->chans[i].rxsig != rxs)
					zt_rbsbits(&sc->chans[i], rxs);
			}
		}
	} else {
		a = __t1_get_reg(sc, 0x60 + x);
		b = __t1_get_reg(sc, 0x63 + x);
		for (y=0;y<8;y++) {
			i = x * 8 + y;
				rxs = 0;
			if (a & (1 << y))
				rxs |= ZT_ABIT;
			if (b & (1 << y))
				rxs |= ZT_BBIT;
			if (!(sc->chans[i].sig & ZT_SIG_CLEAR)) {
				if (sc->chans[i].rxsig != rxs)
					zt_rbsbits(&sc->chans[i], rxs);
			}
		}
	}
}

static void __t1xxp_check_alarms(struct wct1xxp_softc *sc)
{
	unsigned char c,d;
	int alarms;
	int x,j;

	if (sc->ise1) {
		__t1_set_reg(sc, 0x06, 0xff); 
		c = __t1_get_reg(sc, 0x6);
	} else {
		/* Get RIR2 */
		c = __t1_get_reg(sc, 0x31);
		sc->span.rxlevel = c >> 6;
	
		/* Get status register s*/
		__t1_set_reg(sc, 0x20, 0xff); 
		c = __t1_get_reg(sc, 0x20); 
	}

	/* Assume no alarms */
	alarms = 0;

	/* And consider only carrier alarms */
	sc->span.alarms &= (ZT_ALARM_RED | ZT_ALARM_BLUE | ZT_ALARM_NOTOPEN);

	if (sc->ise1) {
		/* XXX Implement me XXX */
	} else {
		/* Detect loopup code if we're not sending one */
		if ((!sc->span.mainttimer) && (c & 0x80)) {
			/* Loop-up code detected */
			if ((sc->loopupcnt++ > 80)  && (sc->span.maintstat != ZT_MAINT_REMOTELOOP)) {
				__t1_set_reg(sc, 0x1e, 0);	/* No local loop */
				__t1_set_reg(sc, 0x0a, 0x40);	/* Remote Loop */
				sc->span.maintstat = ZT_MAINT_REMOTELOOP;
			}
		} else {
			sc->loopupcnt = 0;
		}
		/* Same for loopdown code */
		if ((!sc->span.mainttimer) && (c & 0x40)) {
			/* Loop-down code detected */
			if ((sc->loopdowncnt++ > 80)  && (sc->span.maintstat == ZT_MAINT_REMOTELOOP)) {
				__t1_set_reg(sc, 0x1e, 0);	/* No local loop */
				__t1_set_reg(sc, 0x0a, 0x0);	/* No remote Loop */
				sc->span.maintstat = ZT_MAINT_NONE;
			}
		} else
			sc->loopdowncnt = 0;
	}

	if (sc->span.lineconfig & ZT_CONFIG_NOTOPEN) {
		for (x=0,j=0;x < sc->span.channels;x++)
			if ((sc->chans[x].flags & ZT_FLAG_OPEN) ||
			    (sc->chans[x].flags & ZT_FLAG_NETDEV))
				j++;
		if (!j)
			alarms |= ZT_ALARM_NOTOPEN;
	}

	if (sc->ise1) {
		if (c & 0x9) 
			alarms |= ZT_ALARM_RED;
		if (c & 0x2)
			alarms |= ZT_ALARM_BLUE;
	} else {
		/* Check actual alarm status */
		if (c & 0x3) 
			alarms |= ZT_ALARM_RED;
		if (c & 0x8)
			alarms |= ZT_ALARM_BLUE;
	}
	/* Keep track of recovering */
	if ((!alarms) && sc->span.alarms)
		sc->alarmtimer = ZT_ALARMSETTLE_TIME;

	/* If receiving alarms, go into Yellow alarm state */
	if (alarms && (!sc->span.alarms)) {
#if 0
		printf("Going into yellow alarm\n");
#endif
		if (sc->ise1)
			__t1_set_reg(sc, 0x21, 0x7f); 
		else
			__t1_set_reg(sc, 0x35, 0x11); 
	}

	if (sc->span.alarms != alarms) {
		d = __control_get_reg(sc, WC_CLOCK); 
		start_alarm(sc);
		if (!(alarms & (ZT_ALARM_RED | ZT_ALARM_BLUE | ZT_ALARM_LOOPBACK)) &&
		    sc->sync) {
			/* Use the recieve signalling */
			sc->span.syncsrc = sc->span.spanno;
			d |= 1;
		} else {
			sc->span.syncsrc = 0;
			d &= ~1;
		}
		__control_set_reg(sc, WC_CLOCK, d);  
	}
	if (sc->alarmtimer)
		alarms |= ZT_ALARM_RECOVER;
	if (c & 0x4)
		alarms |= ZT_ALARM_YELLOW;

	sc->span.alarms = alarms;

	zt_alarm_notify(&sc->span);
}

static void __t1xxp_do_counters(struct wct1xxp_softc *sc)
{
	if (sc->alarmtimer) {
		if (!--sc->alarmtimer) {
			sc->span.alarms &= ~(ZT_ALARM_RECOVER);
			/* Clear yellow alarm */
#if 0
			printf("Coming out of alarm\n");
#endif
			if (sc->ise1)
				__t1_set_reg(sc, 0x21, 0x5f);
			else
				__t1_set_reg(sc, 0x35, 0x10);
			zt_alarm_notify(&sc->span);
		}
	}
}

static void wct1xxp_interrupt(void * arg)
{
	struct wct1xxp_softc *sc = arg;
	unsigned char ints;
	int x;

	ints = READ1(WC_INTSTAT);
	WRITE1( WC_INTSTAT, ints);

	if (!ints)
		return;

	if (!sc->intcount) {
		if (debug) printf("Got interrupt: 0x%04x\n", ints);
	}

	sc->intcount++;

	if (sc->clocktimeout && !--sc->clocktimeout) 
		control_set_reg(sc, WC_CLOCK, 0x00 | sc->sync | sc->ise1);

	if (ints & 0x0f) {
		t1xxp_receiveprep(sc, ints);
		t1xxp_transmitprep(sc, ints);
	}

	WCT1_LOCK(sc);

#if 1
	__handle_leds(sc);
#endif

	/* Count down timers */
	__t1xxp_do_counters(sc);

	/* Do some things that we don't have to do very often */
	x = sc->intcount & 15 /* 63 */;
	switch(x) {
	case 0:
	case 1:
	case 2:
		__t1xxp_check_sigbits(sc, x);
		break;
	case 4:
		/* Check alarms 1/4 as frequently */
		if (!(sc->intcount & 0x30))
			__t1xxp_check_alarms(sc);
		break;
	}
	
	WCT1_UNLOCK(sc);

	if (ints & 0x10) 
		printf("PCI Master abort\n");

	if (ints & 0x20)
		printf("PCI Target abort\n");

}

static int t1xxp_hardware_init(struct wct1xxp_softc *sc)
{
	u_int32_t write_dma_addr;
	u_int32_t read_dma_addr;
	unsigned char c;

	/* Hardware PCI stuff */
	/* Reset chip and registers */
	WRITE1( WC_CNTL, DELAY | 0x0e);
	/* Set all outputs to 0 */
	WRITE1( WC_AUXD, 0x00);
	/* Set all to outputs except AUX1 (TDO). */
	WRITE1( WC_AUXC, 0xfd);
	/* Configure the serial port: double clock, 20ns width, no inversion,
	   MSB first */
	WRITE1( WC_SERC, 0xc8);

	/* Internally delay FSC by one */
	WRITE1( WC_FSCDELAY, 0x01);

	/* Back to normal, with automatic DMA wrap around */
	WRITE1( WC_CNTL, DELAY | 0x01);
	
	/* Make sure serial port and DMA are out of reset */
	c =  READ1(WC_CNTL);
	c &= 0xf9;
	WRITE1( WC_CNTL, c);
	
	/* Setup DMA Addresses */
	write_dma_addr = sc->wc_ldata.wc_write_dmaaddr;
	/* Start at writedma */
	WRITE4(WC_DMAWS, write_dma_addr);		/* Write start */
	/* First frame */
	WRITE4(WC_DMAWI, write_dma_addr + ZT_CHUNKSIZE * 32 - 4);		/* Middle (interrupt) */
	/* Second frame */
	WRITE4(WC_DMAWE, write_dma_addr + ZT_CHUNKSIZE * 32 * 2 - 4);			/* End */
	
	read_dma_addr = sc->wc_ldata.wc_read_dmaaddr;
	WRITE4(WC_DMARS, read_dma_addr);	/* Read start */
	/* First frame */
	WRITE4(WC_DMARI, read_dma_addr + ZT_CHUNKSIZE * 32 - 4);	/* Middle (interrupt) */
	/* Second frame */
	WRITE4(WC_DMARE, read_dma_addr + ZT_CHUNKSIZE * 32 * 2 - 4);	/* End */
	
	if (debug) printf("Setting up DMA (write/read = %08lx/%08lx)\n", (long)write_dma_addr, (long)read_dma_addr);
	if (debug) printf("Setting up DMA (write/read = %08lx/%08lx)\n", (long)vtophys(sc->wc_ldata.wc_writechunk), (long)vtophys(sc->wc_ldata.wc_readchunk));

	/* Check out the controller */
	if (debug) printf("Controller version: %02x\n", control_get_reg(sc, WC_VERSION));


	control_set_reg(sc, WC_LEDTEST, 0x00);

	/* Sanity check also determines e1 or t1 */
	if (t1xxp_framer_sanity_check(sc))
		return -1;
	if (sc->ise1)
		sc->chanmap = chanmap_e1;
	else
		sc->chanmap = chanmap_t1;
	/* Setup clock appropriately */
	control_set_reg(sc, WC_CLOCK, 0x02 | sc->sync | sc->ise1);
	sc->clocktimeout = 100;
	
	/* Reset the T1 and report */
	t1xxp_framer_hard_reset(sc);
	start_alarm(sc);
	return 0;

}

static void t1xxp_stop_stuff(struct wct1xxp_softc *sc)
{
	/* Kill clock */
	control_set_reg(sc, WC_CLOCK, 0);

	/* Turn off LED's */
	control_set_reg(sc, WC_LEDTEST, 0);

	/* Reset the T1 */
	t1xxp_framer_hard_reset(sc);

}

/* 
 * syscall routines 
 * We will use them to emulate Linux /proc behaviour --
 * they will give detailed info about card installed
 */ 
#if __FreeBSD_version < 500000
int wct1xxp_sys_open(dev_t dev, int flag, int devtype, struct proc *threadp)
#else
int wct1xxp_sys_open(struct cdev * dev, int flag, int mode, struct thread *threadp)
#endif
{
	return 0;
}

#if __FreeBSD_version < 500000
int wct1xxp_sys_close(dev_t dev, int flags, int fmt, struct proc *threadp)
#else
int wct1xxp_sys_close(struct cdev * dev, int flags, int fmt, struct thread *threadp)
#endif
{
	return 0;
}

#if __FreeBSD_version < 500000
int
wct1xxp_sys_read(dev_t dev, struct uio *uio, int flag)
#else
int
wct1xxp_sys_read(struct cdev *dev, struct uio *uio, int flag)
#endif
{
	int size;
	/* end of file */
	if(uio->uio_offset < 0)
		return -1;

	if(!card_configuration)
		return 0;
	
    if(uio->uio_offset >= strlen(card_configuration))
		return 0;

	size = MIN(uio->uio_resid, strlen(card_configuration) - uio->uio_offset);
	return uiomove(card_configuration +  uio->uio_offset, size, uio);
}


void wct1xxp_sysctl_init()
{
	sysctl_ctx_init(&wct1xxp_clist);

	wct1xxp_root = SYSCTL_ADD_NODE(&wct1xxp_clist, SYSCTL_CHILDREN(zt_root),
		OID_AUTO, "wct1xxp", CTLFLAG_RD, 0, "wct1xxp driver settings");

	SYSCTL_ADD_INT(&wct1xxp_clist, SYSCTL_CHILDREN(wct1xxp_root),
		OID_AUTO, "debug", CTLFLAG_RW, &debug, 0, 
		"make TDM driver more verbose");
}

void wct1xxp_sysctl_destroy()
{
	sysctl_ctx_free(&wct1xxp_clist);
}

void
wct1xxp_dma_map_addr(arg, segs, nseg, error)
		void *arg;
		bus_dma_segment_t *segs;
		int nseg, error;
{
		u_int32_t *paddr;
		
		paddr = arg;
		*paddr = segs->ds_addr;
}



static 
int wct1xxp_dma_allocate(struct wct1xxp_softc * sc)
{
	int error;

	/* write buffers */
#if __FreeBSD_version < 500000
	error = bus_dma_tag_create(NULL, 4, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 32 * 2, 1, ZT_MAX_CHUNKSIZE * 32 * 2, 
			BUS_DMA_ALLOCNOW | BUS_DMA_COHERENT, 
			&sc->wc_ldata.wc_write_tag);
#else
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 32 * 2, 1, ZT_MAX_CHUNKSIZE * 32 * 2, 
			BUS_DMA_ALLOCNOW | BUS_DMA_COHERENT, 
			NULL, NULL, &sc->wc_ldata.wc_write_tag);
#endif

	if (error) {
		printf("wct1xxp: failed to allocate writechunk dma tag\n");
		return 1;
	}

	error = bus_dmamem_alloc(sc->wc_ldata.wc_write_tag,
		(void **)&sc->wc_ldata.wc_writechunk, 
		BUS_DMA_NOWAIT | BUS_DMA_COHERENT | BUS_DMA_ZERO,
		&sc->wc_ldata.wc_write_dmamap);

	if (error) {
		printf("wct1xxp: no memory for writedma buffers!\n");
		/* destroy tag */
		bus_dma_tag_destroy(sc->wc_ldata.wc_write_tag);
		sc->wc_ldata.wc_write_tag = NULL;
		return 1;
	}

	error = bus_dmamap_load(sc->wc_ldata.wc_write_tag,
		sc->wc_ldata.wc_write_dmamap, (void *)sc->wc_ldata.wc_writechunk,
		ZT_MAX_CHUNKSIZE * 32 * 2, wct1xxp_dma_map_addr,
		&sc->wc_ldata.wc_write_dmaaddr, 0);

	if (error) {
			printf("wct1xxp: can't load dma mapping");
			bus_dmamem_free(sc->wc_ldata.wc_write_tag, 
				(void *)sc->wc_ldata.wc_writechunk, 
				sc->wc_ldata.wc_write_dmamap);
			/* destroy map */
			bus_dmamap_destroy(sc->wc_ldata.wc_write_tag, 
				sc->wc_ldata.wc_write_dmamap);
			/* destroy tag */
			bus_dma_tag_destroy(sc->wc_ldata.wc_write_tag);
			sc->wc_ldata.wc_write_tag = NULL;
			return 1;
	}



	/* read buffers */
#if __FreeBSD_version < 500000
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 32 * 2, 1, ZT_MAX_CHUNKSIZE * 32 * 2, 
			BUS_DMA_ALLOCNOW | BUS_DMA_COHERENT, 
			&sc->wc_ldata.wc_read_tag);
#else
	error = bus_dma_tag_create(NULL, 8, 0,
		BUS_SPACE_MAXADDR_32BIT, BUS_SPACE_MAXADDR, NULL, NULL,
		ZT_MAX_CHUNKSIZE * 32 * 2, 1, ZT_MAX_CHUNKSIZE * 32 * 2, 
			BUS_DMA_ALLOCNOW | BUS_DMA_COHERENT, NULL, NULL, &sc->wc_ldata.wc_read_tag);
#endif

	if (error) {
		printf("wct1xxp: failed to allocate readchunk dma tag\n");
		return 1;
	}

	error = bus_dmamem_alloc(sc->wc_ldata.wc_read_tag,
		(void **)&sc->wc_ldata.wc_readchunk, BUS_DMA_NOWAIT | BUS_DMA_COHERENT | BUS_DMA_ZERO,
		&sc->wc_ldata.wc_read_dmamap);

	if (error) {
		printf("wct1xxp: no memory for readdma buffers!\n");
		/* destroy tag */
		bus_dma_tag_destroy(sc->wc_ldata.wc_read_tag);
		sc->wc_ldata.wc_read_tag = NULL;
		return 1;
	}

	error = bus_dmamap_load(sc->wc_ldata.wc_read_tag,
		sc->wc_ldata.wc_read_dmamap, (void *)sc->wc_ldata.wc_readchunk,
		ZT_MAX_CHUNKSIZE * 32 * 2, wct1xxp_dma_map_addr,
		&sc->wc_ldata.wc_read_dmaaddr, BUS_DMA_NOWAIT);

	if (error) {
			printf("wct1xxp: can't load dma mapping");
			bus_dmamem_free(sc->wc_ldata.wc_read_tag, 
				(void *)sc->wc_ldata.wc_readchunk, 
				sc->wc_ldata.wc_read_dmamap);
			/* destroy map */
			bus_dmamap_destroy(sc->wc_ldata.wc_read_tag, 
				sc->wc_ldata.wc_read_dmamap);
			/* destroy tag */
			bus_dma_tag_destroy(sc->wc_ldata.wc_read_tag);
			sc->wc_ldata.wc_read_tag = NULL;
			return 1;
	}

	return 0;
}



/*
 * Return identification string if this is device is ours.
 */
static int
wct1xxp_probe(device_t dev)
{
	uint16_t				vid;
	uint16_t				devid;
	uint16_t				svid;
	struct wct1xxp_ident		*ident;

	vid		= pci_get_vendor(dev);
	devid 	= pci_get_device(dev);
	svid 	= pci_get_subvendor(dev);

	for (ident = wct1xxp_ident_table; ident->device_info != NULL; ident++) {
		if ((ident->vid == vid) && (ident->did == devid) &&
			(ident->svid == svid || ident->svid == -1)) {

			printf("T1/E1 device: vendor=%x device=%x subvendor=%x\n", 
				vid, devid, svid);
			device_set_desc(dev, ident->device_info->name);
			return (0);
		}
	}
	return (ENXIO);
}

/* Attach function is only called if the probe is successful */
static int
wct1xxp_attach(device_t dev)
{
	u_int16_t 				error;
	u_int32_t 				command;
	unsigned int			*canary;
	struct wct1xxp_softc 	* sc;
	int 					unit, rid;
	struct wct1xxp_ident 	*ident;
	uint16_t				vid;
	uint16_t				devid;
	uint16_t				svid;
#if __FreeBSD_version < 500000
	dev_t					d;
#else
	struct cdev *			d;
#endif

	vid		= pci_get_vendor(dev);
	devid 	= pci_get_device(dev);
	svid 	= pci_get_subvendor(dev);
	for (ident = wct1xxp_ident_table; ident->device_info != NULL; ident++)
		if ((ident->vid == vid) && (ident->did == devid) &&
			(ident->svid == svid || ident->svid == -1))
				break;

	if(ident->device_info == 0)
	{
		printf("Card info not found\n");
		error = ENXIO;
		goto fail;
	}
	
	sc = (struct wct1xxp_softc *)device_get_softc(dev);
	unit = device_get_unit(dev);
	/* not realy necessary  */
	memset(sc, 0, sizeof(struct wct1xxp_softc));

	printf("WCT1xxP Attach for wct1xxp%d: deviceID : 0x%x\n",
		unit, pci_get_vendor(dev));

	d = make_dev(&wct1xxp_cdevsw, unit, UID_ROOT,
		GID_WHEEL, 0644, "wct1xxp%d", unit);

	sc->d = d;
	sc->offset = 28;	/* And you thought 42 was the answer */



	/*
	 * Enable bus mastering.
	 */

	pci_enable_busmaster(dev);
	pci_enable_io(dev, SYS_RES_IOPORT);
	pci_enable_io(dev, SYS_RES_MEMORY);
	command = pci_read_config(dev, PCIR_COMMAND, 4);
	// ZAP_LOCK_INIT(sc->lock, "T1/E1 lock");
	WCT1_LOCK(sc->lock);

	if (!(command & PCIM_CMD_PORTEN) && !(command & PCIM_CMD_MEMEN)) {
		printf("wct1xxp%d: failed to enable I/O ports and memory mappings!\n", unit);
		error = ENXIO;
		goto fail;
	}


	sc->wc_ioport_rid = PCIR_BAR(0);
	sc->wc_base_addr = bus_alloc_resource(dev, SYS_RES_IOPORT, 
		&sc->wc_ioport_rid, 0ul, ~0ul, 1, RF_SHAREABLE| RF_ACTIVE);

	if (!sc->wc_base_addr) {
			error = ENOMEM;
			goto fail;
	}
	sc->wc_base_addr_tag	= rman_get_bustag(sc->wc_base_addr);
	sc->wc_base_addr_handle = rman_get_bushandle(sc->wc_base_addr);


	sc->variety = ident->device_info->name;

	/* Alocate memory for two zt chunks - recieve and transmit */

	if(wct1xxp_dma_allocate(sc))
	{
		printf("wct1xxp: Unable to intialize DMA memory\n");
		wct1xxp_detach(dev);
		return ENOMEM;
	}

	/* Initialize canary */
	canary = (volatile unsigned int *)(sc->wc_ldata.wc_readchunk + ZT_CHUNKSIZE * 64 - 4);
	*canary = (CANARY << 16) | (0xffff);

	/* Request IRQ resource */
	rid = 0x0;
	sc->wc_irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid,
		0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);
	if (sc->wc_irq == NULL) {
		printf("IRQ allocation failed!\n");
		error = ENXIO;
		wct1xxp_detach(dev);
		goto fail;
	}

	/* Now we should set up the interrupt handler */

	error = bus_setup_intr(dev, sc->wc_irq, INTR_TYPE_NET | INTR_FAST,
		(void (*)(void *))wct1xxp_interrupt, sc, &(sc->irq_handler));

	if (error) {
		printf("Couldn't set up irq\n");
		wct1xxp_detach(dev);
		goto fail;
	}

	sc->wc_irq_bt = rman_get_bustag(sc->wc_irq);
	sc->wc_irq_bh = rman_get_bushandle(sc->wc_irq);

	/* Initialize hardware */
	t1xxp_hardware_init(sc);

	/* We now know which version of card we have */
	if (sc->ise1)
		sc->variety = "Digium Wildcard E100P E1/PRA";
	else
		sc->variety = "Digium Wildcard T100P T1/PRI";

	/* Misc. software stuff */
	t1xxp_software_init(sc);


	printf("Found a Wildcard: %s\n", sc->variety);

	card_configuration = malloc(80 + 40 + 20, M_T1, M_WAITOK);
	WCT1_UNLOCK(sc->lock);

	return 0;
fail:
	WCT1_UNLOCK(sc->lock);
	printf("We failed: %d\n", error);
	return error;
}

/* Detach device. */
static int
wct1xxp_detach(device_t dev)
{
	struct wct1xxp_softc * sc;

	sc = (struct wct1xxp_softc *)device_get_softc(dev);

	if (sc) {
		/* prevent unloading of module which is using at the moment */
		if(sc->usecount)
			return (EBUSY);

		/* Stop any DMA */
		__t1xxp_stop_dma(sc);

		/* In case hardware is still there */
		__t1xxp_disable_interrupts(sc);
		
		t1xxp_stop_stuff(sc);
		
		/* Immediately free resources */
		bus_generic_detach(dev);

		if (sc->wc_irq)
		{
			bus_teardown_intr(dev, sc->wc_irq, sc->irq_handler);
			bus_release_resource(dev, SYS_RES_IRQ, 0, sc->wc_irq);
		}

		/* Reset PCI chip and registers */
		WRITE1(WC_CNTL, 0x0e);

		if (sc->wc_base_addr != NULL)
			if(bus_release_resource(dev, SYS_RES_IOPORT,
				sc->wc_ioport_rid, sc->wc_base_addr)!= 0)
					printf("Can't release base addr");


		if (sc->wc_ldata.wc_read_tag) {
			bus_dmamap_unload(sc->wc_ldata.wc_read_tag,
				sc->wc_ldata.wc_read_dmamap);
			bus_dmamem_free(sc->wc_ldata.wc_read_tag, 
				(void *)sc->wc_ldata.wc_readchunk,
				sc->wc_ldata.wc_read_dmamap);
			bus_dma_tag_destroy(sc->wc_ldata.wc_read_tag);
		}

		if (sc->wc_ldata.wc_write_tag) {
			bus_dmamap_unload(sc->wc_ldata.wc_write_tag,
				sc->wc_ldata.wc_write_dmamap);
			bus_dmamem_free(sc->wc_ldata.wc_write_tag, 
				(void *)sc->wc_ldata.wc_writechunk,
				sc->wc_ldata.wc_write_dmamap);
			bus_dma_tag_destroy(sc->wc_ldata.wc_write_tag);
		}

		if(sc->d)
			destroy_dev(sc->d);

		if (!sc->usecount)
			t1xxp_release(sc);
		else
			sc->dead = 1;

		if(card_configuration)
		{
			free(card_configuration, M_T1);
			card_configuration = 0;
		}
		printf("WCT1 detach!\n");
	}
	// ZAP_LOCK_DESTROY(sc->lock);
	return 0;
}




/* Called during system shutdown after sync. */
static int
wct1xxp_shutdown(device_t dev)
{
	return (0);
}

/*
 * Device suspend routine.
 */
static int
wct1xxp_suspend(device_t dev)
{
	return (0);
}

/*
 * Device resume routine.
 */

static int
wct1xxp_resume(device_t dev)
{
	return (0);
}




static int
wct1xxp_load(module_t mod, int cmd, void *arg)
{
    int err = 0;
	switch (cmd) {
		case MOD_LOAD:
    		err = driver_module_handler(mod, cmd, arg);
			// if(!err)
				// wct1xxp_sysctl_init();
		break;          /* Success*/
     	
		case MOD_UNLOAD:
    		err = driver_module_handler(mod, cmd, arg);
			// if(!err)
				// wct1xxp_sysctl_destroy();
		break;          /* Success */
 
   		default: 
			err = EINVAL;
 		break;
	}
	return err;
}



static device_method_t wct1xxp_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,	 	wct1xxp_probe),
	DEVMETHOD(device_attach,	wct1xxp_attach),
	DEVMETHOD(device_detach,	wct1xxp_detach),
	DEVMETHOD(device_shutdown,  wct1xxp_shutdown),
	DEVMETHOD(device_suspend,   wct1xxp_suspend),
	DEVMETHOD(device_resume,	wct1xxp_resume),

	{ 0, 0 }
};

static driver_t wct1xxp_driver = {
	"wct1xxp",
	wct1xxp_methods,
	sizeof(struct wct1xxp_softc)
};

static devclass_t wct1xxp_devclass;

#if __FreeBSD_version < 500000
static driver_t *wct1xxp_pci_driver_list = { &wct1xxp_driver };
static struct driver_module_data wct1xxp_pci_driver_mod = { 
	0, 0,
	"pci",
	&wct1xxp_pci_driver_list,
	sizeof(wct1xxp_pci_driver_list)/sizeof(wct1xxp_pci_driver_list[0]),
	&wct1xxp_devclass
};

static moduledata_t wct1xxp_pci_mod = {
        "pci/wct1xxp",
        wct1xxp_load,
        &wct1xxp_pci_driver_mod
};

#else

static struct driver_module_data wct1xxp_pci_driver_mod = { 
	0, 0,
	"pci",
	(kobj_class_t) &wct1xxp_driver,
	&wct1xxp_devclass
};

static moduledata_t wct1xxp_pci_mod = {
        "pci/wct1xxp",
        wct1xxp_load,
        &wct1xxp_pci_driver_mod
};
#endif

MODULE_DEPEND(wct1xxp, zaptel, 1, 1, 1);
DECLARE_MODULE(wct1xxp, wct1xxp_pci_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);

